"""

"""
from numpy import zeros, copy, where, dot, hstack, argsort, sum
from numpy.linalg import det
from scipy.linalg import eigh
from os.path import expanduser, expandvars

from DFTB import utils
from DFTB.DFTB2 import DFTB2
from DFTB.AtomicData import hartree_to_eV
from DFTB import XYZ

class ConstrainedCI(DFTB2):
    CIconstraints = []
    def loadConstraints(self, constraints_script=None):
        """
        read constraints for CI calculation from a python file. In the main
        body the variable "CIconstraints" has to be defined, which contains
        a list of constraints, one for each diabatic state.
        For example,

           constraintsA = [([0,1],-1.0)]
           constraintsB = [([0,1],0.0)]
           CI_constraints = [constraintsA, constraintsB]
           
        defines two diabatic states A and B.

        Parameters:
        ===========
        constraints_script: path to script with CI constraints
        """
        if constraints_script == None:
            raise Exception("Cannot run constrained CI without constraints script!")
        dic = {}
        try:
            execfile(expanduser(expandvars(constraints_script)), dic)
        except IOError as e:
            print "Cannot load constraints"
            print e
        self.CIconstraints = dic.get("CIconstraints", [])
        if self.CIconstraints == []:
            raise Exception("Variable CIconstraints is empty or not defined in '%s'!" % constraints_script)
    def getEnergy(self, calc_gradients=False, **opts):
        # construct H0 and S matrices
        # they stay the same for all diabatic states
        self.S, self.H0 = self._constructH0andS()
        self.gamma = self._construct_gamma()
        # perform constrained SCC calculation for each set of constraints
        # and store MO coefficients, energies and constraining potentials for each
        # diabatic state.
        entot_diab = []
        f_diab = [] # occupation numbers
        orbs_diab = []
        Vc_diab = []
        engrad_diab = [] # diabatic energy gradients
        orbs_grad_diab = [] # gradients of MO coefficients
        for idiab, constraints in enumerate(self.CIconstraints):
            print "Diabatic State %s" % idiab
            print "================="
            print "Constraints: %s" % constraints
            self._setConstraints(constraints)
            self.runConstrainedSCC(**opts)
            # 
            Etot = self.getEnergies()[-1]
            entot_diab.append(Etot)
            f_diab.append(copy(self.f))
            orbs_diab.append(copy(self.orbs))
            Vc_diab.append(copy(self.Vc))
            if calc_gradients == True:
                # energy gradient
                engrad = super(ConstrainedCI, self).getGradient(format="")
                engrad_diab.append(grad)
                # gradient of MO coefficients
                orbs_grad = self.solveCPE()
                orbs_grad_diab.append(orbs_grad)
            #
            from DFTB.Molden import MoldenExporter
            molden = MoldenExporter(self)
            molden.export(molden_file="/tmp/molden_diabatic_%d.input" % idiab)
            #
            print "TOTAL DIABATIC ENERGY              : %s %s\n" % (string.rjust("%.7f hartree" % self.E_tot,20), string.rjust("%.7f eV" % (self.E_tot*hartree_to_eV),20))
        print entot_diab
        CImat, olap, vcI = self._buildCImatrix(entot_diab, f_diab, orbs_diab, Vc_diab, engrad_diab, orbs_grad_diab)
        # transform CI matrix into the basis of the eigenvectors of vcI matrix
        # see J. Chem. Phys. 140, 104105 (2014) eq. 13
        L,V =eigh(vcI, olap)
        Hci = dot(V.transpose(), dot(CImat, V))
        print "matrix H in the orthogonalized basis (in eV):"
        print "============================================="
        dim,dim = Hci.shape
        print utils.annotated_matrix(Hci*hartree_to_eV, ["%.2d" % i for i in range(0, dim)], ["%.2d" % j for j in range(0, dim)])
        # diagonalize H matrix
        En,CIcoefs = eigh(Hci, olap)
        sort_indx = argsort(En)
        En_sorted = En[sort_indx]
        CIcoefs_sorted = CIcoefs[:, sort_indx]
        print "Adiabatic Energies:"
        print "==================="
        for I,EnI in enumerate(En_sorted):
            print "adiabatic energy %d = %.7f Hartree" % (I, EnI)
        print "CI Coefficients:"
        print "================"
        for I,EnI in enumerate(En_sorted):
            print "%d: %s" % (I, CIcoefs_sorted[:,I])

        return En_sorted
    def getGradient(self, format="atomlist"):
        pass
    def _energy_func(self, x, atomlist, state=0):
        """helper function for numerical differentiation"""
        atpos = XYZ.vector2atomlist(x, atomlist)
        self.setGeometry(atpos, keep_charge=True)
        En_adiab = self.getEnergy(maxiter=2000, scc_conv=1.0e-5, constr_conv=1.0e-5, DIIS_memory=2, search_bracket=(-10.0, 10.0))
        return En_adiab[state]
    def getGradientNumerically(self, atomlist, state=0, format="atomlist"): # for testing
        h = 1.0e-10
        x0 = XYZ.atomlist2vector(atomlist)
        # compute partial derivatives dE0/dx_i, dE0/dy_i dE0/dz_i as difference quotients
        dEdx = utils.numerical_gradient(lambda x: self._energy_func(x,atomlist,0), x0)
        # check that the sum of the forces is zero
        Fxtot = sum(dEdx[0::3])
        Fytot = sum(dEdx[1::3])
        Fztot = sum(dEdx[2::3])
        print "dEdx"
        print dEdx
        if self.verbose > -1:
            print "Total forces from finite difference: Fx = %2.6f  Fy = %2.6f  Fz = %2.6f" % (Fxtot, Fytot, Fztot)
        tol = 1.0e-3
        assert abs(Fxtot) < tol and abs(Fytot) < tol and abs(Fztot) < tol
        #
        self.setGeometry(atomlist) # restore geometry
        if format == "atomlist":
            grad = XYZ.vector2atomlist(dEdx, atomlist)
        else:
            grad = dEdx
        return grad
    def _buildCImatrix(self, entot_diab, f_diab, orbs_diab, Vc_diab, engrad_diab, orbs_grad_diab):
        dim = len(entot_diab)
        # overlap between diabatic state. Although there is no wavefunction
        # in DFT, the diabatic states are approximated by the KS Slater determinants.

        # Which orbitals are occupied in each diabatic state?
        # Because we will usually have unpaired electrons, we need to
        # distinguish between spin-up and spin-down orbitals
        occa_orbs_diab = []
        occb_orbs_diab = []
        for i in range(0, dim):
            # doubly occupied orbitals
            occ_double_indx = where(f_diab[i] > 1.0)[0]
            # singly occupied orbitals
            occ_single_indx = where((1.0 >= f_diab[i]) & (f_diab[i] > 0.0))[0]
            print "Doubly occupied orbitals: %s" % occ_double_indx
            print "Singly occupied orbitals: %s" % occ_single_indx

            # occupied spin-up orbitals -> doubly occupied orbitals
            occa_indx = occ_double_indx
            # occupied spin-down orbitals -> doubly occupied + singly occupied
            # so the unpaired electrons are arbitrarily assigned to be spin down
            # If more than 1 electron is unpaired we always get the maximum spin
            # although other multiplicities are also possible.
            occb_indx = hstack((occ_double_indx, occ_single_indx))
            # mo coefficients of spin orbitals
            occa_orbs_diab.append( orbs_diab[i][:,occa_indx] )
            occb_orbs_diab.append( orbs_diab[i][:,occb_indx] )

        # The overlap between two Slater determinants is the Slater determinant of the
        # overlaps.
        olap = zeros((dim,dim))  # <i|j>
        for i in range(0, dim):
            olap[i,i] = 1.0
            for j in range(i,dim):
                # orbsa_diab[i][:,a] are the MO coefficients of the i-th spin-up MO
                # in the Slater determinant. The Slater determinant has two spin
                # blocks
                olapa = det( dot(occa_orbs_diab[i].transpose(), dot(self.S, occa_orbs_diab[j])) ) 
                olapb = det( dot(occb_orbs_diab[i].transpose(), dot(self.S, occb_orbs_diab[j])) ) 
                print "i = %s  j = %s" % (i,j)
                print "olapa = %s" % olapa
                print "olapb = %s" % olapb
                olap[i,j] = olapa*olapb
                olap[j,i] = olap[i,j]
        print "Diabatic Overlaps <i|j>:"
        print "========================"
        print utils.annotated_matrix(olap, ["%.2d" % i for i in range(0, dim)], ["%.2d" % j for j in range(0, dim)])
        # matrix elements of contraints potential
        vcI = zeros((dim,dim))    # <i|wI(r)|j> = <i|P^I|j>
        for i in range(0, dim):
#            print "Vc_diab[%d] = %s" % (i, Vc_diab[i].tolist())
            for j in range(0, dim):
                print "i,j = %d,%d" % (i,j)
#                vcIa = det( dot(occa_orbs_diab[i].transpose(), dot(Vc_diab[i], occa_orbs_diab[j])) )
#                vcIb = det( dot(occb_orbs_diab[i].transpose(), dot(Vc_diab[i], occb_orbs_diab[j])) )
                vcIa = sum( dot(occa_orbs_diab[i].transpose(), dot(Vc_diab[i], occa_orbs_diab[j])) )
                vcIb = sum( dot(occb_orbs_diab[i].transpose(), dot(Vc_diab[i], occb_orbs_diab[j])) )
                print "vcIa = %s" % vcIa
                print "vcIb = %s" % vcIb
                vcI[i,j] = vcIa*vcIb
#                print "vcIa = det(%s)" % ( dot(occa_orbs_diab[i].transpose(), dot(Vc_diab[i], occa_orbs_diab[j])) )
#                print "vcIb = det(%s)" % ( dot(occb_orbs_diab[i].transpose(), dot(Vc_diab[i], occb_orbs_diab[j])) )

        print "Constraining potentials <i|Vc^i(r)|j>:"
        print "===================================="
        print utils.annotated_matrix(vcI, ["%.2d" % i for i in range(0, dim)], ["%.2d" % j for j in range(0, dim)], format="%.30f")
        # build CI matrix
        CImat = zeros((dim,dim))
        for i in range(0, dim):
            # diagonal elements
            CImat[i,i] = entot_diab[i]
            for j in range(i+1,dim):
                CImat[i,j] = 0.5*( \
                    (entot_diab[i] + entot_diab[j])*olap[i,j]
                   -vcI[i,j] - vcI[j,i] )
                CImat[j,i] = CImat[i,j]
        print "CI matrix <i|H|j> (in eV):"
        print "=========================="
        print utils.annotated_matrix(CImat*hartree_to_eV, ["%.2d" % i for i in range(0, dim)], ["%.2d" % j for j in range(0, dim)])
        """
        # GRADIENTS
        # Gradients of constraining potentials
        
        # Gradients of CI matrix elements
        nR = 3*len(self.atomlist)
        CImat_grad = zeros((dim,dim,nR))
        for i in range(0, dim):
            # diagonal elements are gradients of diabatic energies
            CImat_grad[i,i,:] = engrad_diab[i]
            for j in range(i+1,dim):
                olapij_grad = 0
                CImat_grad[i,j,:] = \
                    0.5 * olap[i,j] * (engrad_diab[i] + engrad_diab[j]) \
                  + 0.5 * (entot_diab[i] + entot_diab[j]) * olapij_grad \
                  - 0.5*(vcI_grad[i,j] + vcI_grad[j,i])
                CImat_grad[j,i,:] = CImat_grad[i,j,:]
        """
        return CImat, olap, vcI


if __name__ == "__main__":
    from DFTB.XYZ import read_xyz, extract_keywords_xyz
    import sys
    import os.path
    import string
    usage = "Usage: %s <xyz-file>" % sys.argv[0]
    if len(sys.argv) < 2:
        print usage
        exit(-1)
    xyz_file = sys.argv[1]
    # read first structure to get the atom types
    structures = read_xyz(xyz_file)
    atomlist = structures[0]
    kwds = extract_keywords_xyz(xyz_file)
    parser = utils.OptionParserFuncWrapper(ConstrainedCI([]).loadConstraints, usage)
    (options, args) = parser.parse_args()
    if options.get("constraints_script", None) == None:
        raise Exception("You have to specify a constraints script with the --constraints_script option!")

#    CI = ConstrainedCI(atomlist, parameter_set="hotbit", verbose=0, **options)
    CI = ConstrainedCI(atomlist, parameter_set="homegrown", verbose=0, **options)
#    CI = ConstrainedCI(atomlist, parameter_set="3ob", verbose=0, **options)
    for i,atomlist in enumerate(structures):
        CI.setGeometry(atomlist, kwds.get("charge", 0.0))
        En_adiab = CI.getEnergy(maxiter=2000, scc_conv=1.0e-5, constr_conv=1.0e-5, DIIS_memory=5, search_bracket=(-10.0, 10.0))
        print "ADIABATIC ENERGY OF STRUCTURE %d              : %s %s\n" % (i, string.rjust("%.7f hartree" % En_adiab[0],20), string.rjust("%.7f eV" % (En_adiab[0]*hartree_to_eV),20))

#        print CI.getGradientNumerically(atomlist)
