#!/usr/bin/env python
"""
Optimize the geometry of a molecular crystal using a periodic force field.
"""
import numpy
import numpy.linalg as la
import os.path

from DFTB import XYZ
from DFTB.ForceField.PeriodicForceField import read_force_field, PeriodicForceField, enlarged_unitcell

if __name__ == "__main__":
    import sys
    from scipy import optimize
    import optparse

    usage  = "Usage: %s <.ff force field file>\n" % os.path.basename(sys.argv[0])
    usage += "  optimize geometry using periodic force field\n"
    usage += "  Force field files can be found in the folder DFTB/ForceField/examples/\n"
    usage += "  type --help to see all options.\n"

    parser = optparse.OptionParser(usage)
    parser.add_option("--nr_unit_cells", dest="nr_unit_cells", type=int, default=0, help="In the output the N neighbouring unit cells are added to the geometry to give an impression of a larger block of the crystal. [default: %default]")
    
    (opts, args) = parser.parse_args()
    if len(args) < 1:
        print usage
        exit(-1)

    
    ff_file = args[0]  #"h2.ff" #"ethene.ff" #"pyrene_crystal_expanded.ff" #
    
    atomlist, atomtypes, lattice_vectors = read_force_field(ff_file)
    force_field = PeriodicForceField(atomlist, atomtypes, lattice_vectors)

    # convert geometry to a vector
    x0 = XYZ.atomlist2vector(atomlist)

    # FIND ENERGY MINIMUM
    # f is the objective function that should be minimized
    # it returns (f(x), f'(x))
    def f(x):
        # also save geometries from line searches
        save_xyz(x)
        #
        energy, grad = force_field.getEnergyAndGradient(x)
        print "E = %2.7f     |grad| = %2.7f" % (energy, la.norm(grad))
        #
        return energy, grad

    #xyz_opt = os.path.join("/tmp/", os.path.basename(ff_file).replace(".ff", "_opt.xyz"))
    xyz_opt = ff_file.replace(".ff", "_opt.xyz")
    print "Intermediate geometries will be written to %s" % xyz_opt
    # This is a callback function that is executed by numpy for each optimization step.
    # It appends the current geometry to an xyz-file.
    def save_xyz(x, mode="a"):
        atomlist_opt = XYZ.vector2atomlist(x, atomlist)
        atomlist_opt = enlarged_unitcell(atomlist_opt, lattice_vectors, nmax=opts.nr_unit_cells)
        XYZ.write_xyz(xyz_opt, [atomlist_opt], \
                      mode=mode)
    # save initial geometry
    save_xyz(x0, mode="w")
        
    options = {'gtol': 1.0e-7, 'norm': 2}
    # The "BFGS" method is probably better than "CG", but the line search in BFGS is expensive.
    res = optimize.minimize(f, x0, method="CG", jac=True, callback=save_xyz, options=options)
    # save optimized geometry
    xopt = res.x
    save_xyz(xopt)
    print "Optimized geometry written to %s" % xyz_opt
